'use strict';

var _require = require('./constants');

const REQUIRE = _require.REQUIRE,
      MODULE = _require.MODULE,
      EXPORTS = _require.EXPORTS,
      DEFINE = _require.DEFINE,
      AMD_DEFINE_RESULT = _require.AMD_DEFINE_RESULT;

const createHelpers = require('./helpers');

module.exports = ({ types: t }) => {
  var _createHelpers = createHelpers({ types: t });

  const decodeDefineArguments = _createHelpers.decodeDefineArguments,
        decodeRequireArguments = _createHelpers.decodeRequireArguments,
        isModuleOrExportsInjected = _createHelpers.isModuleOrExportsInjected,
        createRequireExpression = _createHelpers.createRequireExpression,
        createModuleExportsAssignmentExpression = _createHelpers.createModuleExportsAssignmentExpression,
        createModuleExportsResultCheck = _createHelpers.createModuleExportsResultCheck;


  const argumentDecoders = {
    [DEFINE]: decodeDefineArguments,
    [REQUIRE]: decodeRequireArguments
  };

  // Simple version of zip that only pairs elements until the end of the first array
  const zip = (array1, array2) => {
    return array1.map((element, index) => [element, array2[index]]);
  };

  const ExpressionStatement = path => {
    const node = path.node,
          parent = path.parent;


    if (!t.isCallExpression(node.expression)) return;

    const name = node.expression.callee.name;

    const isDefineCall = name === DEFINE;

    if (isDefineCall && !t.isProgram(parent)) return;

    const argumentDecoder = argumentDecoders[name];

    if (!argumentDecoder) return;

    var _argumentDecoder = argumentDecoder(node.expression.arguments);

    const dependencyList = _argumentDecoder.dependencyList,
          factory = _argumentDecoder.factory;


    if (!t.isArrayExpression(dependencyList) && !factory) return;

    const isFunctionFactory = t.isFunctionExpression(factory);
    const requireExpressions = [];
    // Order is important here for the simplified commonjs wrapper
    const keywords = [REQUIRE, EXPORTS, MODULE];

    if (dependencyList) {
      const dependencyParameterPairs = zip(dependencyList.elements, isFunctionFactory ? factory.params : []);

      const explicitRequires = dependencyParameterPairs.filter(([dependency]) => {
        return !t.isStringLiteral(dependency) || keywords.indexOf(dependency.value) === -1;
      }).map(([dependency, paramName]) => {
        return createRequireExpression(dependency, paramName);
      });

      requireExpressions.push(...explicitRequires);
    }

    if (isFunctionFactory) {
      const factoryArity = factory.params.length;
      let replacementFuncExpr = t.functionExpression(null, [], t.blockStatement(requireExpressions.concat(factory.body.body)));
      let replacementCallExprParams = [];

      // https://github.com/requirejs/requirejs/wiki/differences-between-the-simplified-commonjs-wrapper-and-standard-amd-define
      const isSimplifiedCommonJSWrapper = !dependencyList && factoryArity > 0;
      if (isSimplifiedCommonJSWrapper) {
        replacementFuncExpr = factory;
        replacementCallExprParams = keywords.slice(0, factoryArity).map(a => t.identifier(a));
      }

      const factoryReplacement = t.callExpression(replacementFuncExpr, replacementCallExprParams);

      if (isDefineCall) {
        if (!isModuleOrExportsInjected(dependencyList, factoryArity)) {
          path.replaceWith(createModuleExportsAssignmentExpression(factoryReplacement));
        } else {
          const resultCheckIdentifier = path.scope.hasOwnBinding(AMD_DEFINE_RESULT) ? path.scope.generateUidIdentifier(AMD_DEFINE_RESULT) : t.identifier(AMD_DEFINE_RESULT);
          path.replaceWithMultiple(createModuleExportsResultCheck(factoryReplacement, resultCheckIdentifier));
        }
      } else {
        path.replaceWith(factoryReplacement);
      }
    } else if (factory && isDefineCall) {
      const exportExpression = createModuleExportsAssignmentExpression(factory);
      const nodes = requireExpressions.concat(exportExpression);
      path.replaceWithMultiple(nodes);
    } else {
      path.replaceWithMultiple(requireExpressions);
    }
  };

  return {
    visitor: {
      ExpressionStatement
    }
  };
};